/* Compute the characteristic path of a call:
  This version computes just the immediate choices as
  the characteristic path */

/* JPG: 1.8.90 */


charpath(first_call(R),P1) :-
	!,
	functor(R,P,N),
	name(P,Plist),
	name('**',Star),
	my_append(Plist,Star,Pstarlist),
	name(Pstar,Pstarlist),
	P1 =.. [Pstar,N].
charpath(R,P1) :-
	functor(R,P,N),
	P1 =.. [P,N].

resultant(first_call(G),R) :-
	!,
	first_call_clause(G,R).
resultant(G,R) :-
	first_call_clause(G,R).

first_call_clause(G,R) :-
	clause(G,R1),
	first_call_optimise(G,R1,R).

first_call_optimise(G,R1,R) :-
	first_call(R1,B),
	nonvar(B),
	functor(G,P,N),
	functor(B,P,N),
	!,
	more_specific(R1,R).
first_call_optimise(_,R,R).

cleanup(_,Cls,Cls).

first_call(B,B) :-
	var(B),
	!.
first_call((B,_),A) :-
	!,
	first_call(B,A).
first_call(B,B).

more_specific(B,B) :-
	var(B),
	!.
more_specific((B,Bs),(A,Bs)) :-
	!,
	more_specific(B,A).
more_specific(B,B) :-
        sp_builtin(B),
        unfoldable(B),
        !,
        B.
more_specific(B,B) :-
        sp_builtin(B),
        !.
more_specific(B,B) :-
        imported(B),
        !.
more_specific(B,first_call(B)) :-
        solve_choice(B,_).

more_specific(_).

