/* read in a program, for Sicstus version of SP system */
/* use of this procedure avoids the need to 
   declare each predicate as dynamic */

:- dynamic(program_preds/1).
:- dynamic(program_pred/2).
:- dynamic(directive/1).
:- dynamic(ub_clause/2).

sp_consult(F) :-
	sp_ttynl,
	see(F),
	abolish_all_existing,
	read(C),
	read_clauses(C,Ps,K),
	seen,
	sp_ttynl,
	write(stdout,'Finished reading '),write(stdout,F),
	sp_ttynl,
	write(stdout,'Number of clauses: '),
	write(stdout,K),
	sp_ttynl,
	assert(program_preds(Ps)),
	strong_connected_components(Ps,SCCs),
        put_sccs(SCCs),
	!.
sp_consult(F) :-
	write(stdout,'Cannot find file '), write(stdout,F),
	write(stdout,', or problems while reading.'), 
	seen,
	assert(program_preds([])),
	sp_ttynl,sp_ttynl.

read_clauses(C,Ps,0) :- 
	end_of_file(C),
	!,
	close_list(Ps).
read_clauses((:- entry(G,_)),Ps,N) :-
        !,
        assert(directive(entry(G))),
        read(C1),
        read_clauses(C1,Ps,N).
read_clauses((:- import(P/N)),Ps,K) :-
        !,
        assert(directive(import(P/N))),
        functor(H,P,N),
        assert(ub_clause(imported(H),true)),
        read(C1),
        read_clauses(C1,Ps,K).
read_clauses((:- import(from(P/N,M))),Ps,K) :-
        !,
        assert(directive(import(from(P/N,M)))),
        functor(H,P,N),
        assert(ub_clause(imported(H),true)),
        read(C1),
        read_clauses(C1,Ps,K).
read_clauses((:- dynamic(P/N)),Ps,K) :-
        !,
        assert(directive(dynamic(P/N))),
        functor(H,P,N),
        assert(ub_clause(imported(H),true)),
        read(C1),
        read_clauses(C1,Ps,K).
read_clauses((:- D),Ps,K) :-
        memb2(D,[global(_),index(_,_),setdebug]),
        !,
        assert(directive(D)),
        read(C1),
        read_clauses(C1,Ps,K).
read_clauses((:- Dir),Ps,K) :- 
	!,
	call(Dir),
	assert(directive(Dir)),
	read(C1),
	!,
	read_clauses(C1,Ps,K).
read_clauses(C,Ps,K) :- 
	get_pred_name(C,Pred,Bodypreds),
        each_memb1([Pred|Bodypreds],Ps),
        store_clause(C),
        read(C1),
        !,
        read_clauses(C1,Ps,K1),
	K is K1+1.

get_pred_name((H :- B),P/N,BPs) :-
        !,
        functor(H,P,N),
        body_preds(B,BPs).
get_pred_name(H ,P/N,[]) :-
        functor(H,P,N).

body_preds(true,[]) :-
        !.
body_preds((\+ B,Bs),Ps) :-
        !,
        body_preds((B,Bs),Ps).
body_preds((B,Bs),Ps) :-
        sp_builtin(B),
        !,
        body_preds(Bs,Ps).
body_preds((B,Bs),[P/N|Ps]) :-
        !,
        functor(B,P,N),
        body_preds(Bs,Ps).
body_preds(\+ B,Ps) :-
        !,
        body_preds(B,Ps).
body_preds(B,[]) :-
        sp_builtin(B),
        !.
body_preds(B,[P/N]) :-
        functor(B,P,N).

each_memb1([],_).
each_memb1([P|Ps],S) :-
        memb1(P,S),
        each_memb1(Ps,S).


close_list([]) :-
	!.
close_list([_|X]) :-
	close_list(X).

abolish_all_existing :-
%	program_preds(Ps),
	!,
	write(stdout,'Clearing workspace....'),
%	abolish_each(Ps),
	abolish(ub_clause(_,_)),
	abolish(directive(_)),
	assert(directive(dummy)),
	abolish(program_preds(_)),
	abolish(program_pred(_,_)).
abolish_all_existing.

abolish_each([]) :-
	nl.
abolish_each([P/N|Ps]) :-
	functor(H,P,N),
	abolish(H),
	!,
	abolish_each(Ps).


put_sccs([]).
put_sccs([(R,C)|Cs]) :-
        put_scc(R,C),
        put_sccs(Cs).

put_scc(_,[]).
put_scc(R,[P|Ps]) :-
        assert(program_pred(P,R)),
        put_scc(R,Ps).

store_clause((H :- B)) :-
        !,
        assert(ub_clause(H,B)).
store_clause(H) :-
        assert(ub_clause(H,true)).

