ans_clauses(X) :- 
	end_of_file(X),
	!.
ans_clauses((A :- B)) :- 
	!,
	canonical((A :- B)),
	answerlits(B,Bs),
	flat_name('_ans'(A),AA),
	flat_name(query(A),QA),
	flat_names(Bs,FBs),
	displayqRClauses([(AA :- QA,FBs)]),
	read(C1),
	!,
	ans_clauses(C1).
ans_clauses(A) :- 
        canonical(A),
	flat_name('_ans'(A),AA),
        flat_name(query(A),QA),
        displayqRClauses([(AA :- QA)]),
	read(C1),
	!,
	ans_clauses(C1).

answerlits((B,Bs),(B,Bs1)) :-
	sp_builtin(B),
	!,
	answerlits(Bs,Bs1).
answerlits((B,Bs),('_ans'(B),Bs1)) :-
	!,
	answerlits(Bs,Bs1).
answerlits(B,B) :-
	sp_builtin(B).
answerlits(B,'_ans'(B)).

flat_name(A,A) :-
	sp_builtin(A),
	!.
flat_name(A,A1) :-
	A =.. [F,As],
	As =.. [G|Args],
	name(F,Fn),
	name(G,Gn),
	concat(Gn,[95|Fn],An),
	name(H,An),
	A1 =.. [H|Args].

flat_names((B,Bs),(B1,Bs1)) :-
	!,
	flat_name(B,B1),
	flat_names(Bs,Bs1).
flat_names(B,B1) :-
	flat_name(B,B1).

magic(F,F1,Q) :-
	sp_consult(F),
	write('computing dependency graph...'),
	program_preds(Ps),
	strong_connected_components(Ps,PG),
	new_line,
	write('generating transformed program...'),
	see(F),
	tell(F1),
	read(C),
	ans_clauses(C),
	seen,
	see(F),
	read(C1),
	query_clauses(C1,PG),
	flat_name(query(Q),QQ),
	canonical( QQ ),
	displayqRClauses([(QQ :- true)]),
	nl,
	seen,
	see(F),
	read(C2),
	orig_clauses(C2),
	nl,
	seen,
	told,
	new_line.

query_clauses(Z,_) :- 
	end_of_file(Z),
	!.
query_clauses((A :- B),PG) :- 
	!,
	canonical((A :- B)),
	conj_to_list(B,Lits),
	bodylitqueries(Lits,1,A,Lits,PG),
	read(C1),
	!,
	query_clauses(C1,PG).
query_clauses(_,PG) :- 
	read(C1),
	!,
	query_clauses(C1,PG).

bodylitqueries([],_,_,_,_) :-
	!.
bodylitqueries([Bj|Bs],J,A,Lits,PG) :-
	sp_builtin(Bj),
	!,
	J1 is J+1,
	bodylitqueries(Bs,J1,A,Lits,PG).
bodylitqueries([Bj|Bs],J,A,Lits,PG) :-
	!,
	remove_jth(J,Lits,Bj,Lits1),
	answerlits(Lits1,1,J,Bs2,Bs3,Bj,PG),
	flat_names(Bs2,QBs2),
	joingoals(QBs2,Bs3,QBs),
	flat_name(query(Bj),QBj),
	flat_name(query(A),QA),
	displayqRClauses([(QBj :- QA,QBs)]),
	J1 is J+1,
	bodylitqueries(Bs,J1,A,Lits,PG).

remove_jth(1,[B|Bs],B,[dummy|Bs]) :-
	!.
remove_jth(J,[B|Bs],B1,[B|Bs1]) :-
	J1 is J-1,
	remove_jth(J1,Bs,B1,Bs1).

answerlits([],_,_,true,true,_,_) :-
	!.
answerlits([dummy],_,_,true,true,_,_) :-
	!.
answerlits([B],_,_,true,B,_,_) :-
	sp_builtin(B),
	!.
answerlits([B],I,J,'_ans'(B),true,Bj,PG) :-
	depends_on(Bj,B,I,J,PG),
	!.
answerlits([B],_,_,true,B,_,_) :-
	!.
answerlits([dummy|Bs],I,J,Bs2,Bs3,Bj,PG) :-
	!,
	I1 is I+1,
	answerlits(Bs,I1,J,Bs2,Bs3,Bj,PG).
answerlits([B|Bs],I,J,Bs2,(B,Bs3),Bj,PG) :-
	sp_builtin(B),
	!,
	I1 is I+1,
	answerlits(Bs,I1,J,Bs2,Bs3,Bj,PG).
answerlits([B|Bs],I,J,('_ans'(B),Bs2),Bs3,Bj,PG) :-
	depends_on(Bj,B,I,J,PG),
	!,
	I1 is I+1,
	answerlits(Bs,I1,J,Bs2,Bs3,Bj,PG).
answerlits([B|Bs],I,J,Bs2,(B,Bs3),Bj,PG) :-
	I1 is I+1,
	answerlits(Bs,I1,J,Bs2,Bs3,Bj,PG).

depends_on(B1,B2,I,J,PG) :-
        functor(B1,P,N),
        functor(B2,Q,M),
        depends_on1(P/N,Q/M,I,J,PG).

depends_on1(P,Q,I,J,[(_,Ds)|PG]) :-
        \+ memb1(Q,Ds),
        !,
        depends_on2(P,Q,I,J,Ds,PG).
depends_on1(P,_,I,J,[(_,Ds)|_]) :-
        memb1(P,Ds),
        I > J.                  % mutually recursive preds.

depends_on2(P,_,_,_,Ds,_) :-
        memb1(P,Ds),
        !.
depends_on2(P,Q,I,J,_,PG) :-
        depends_on1(P,Q,I,J,PG).


concat([],Xs,Xs).
concat([X|Xs],Ys,[X|Zs]) :-
	concat(Xs,Ys,Zs).

conj_to_list(true,[]) :-
	!.
conj_to_list((B,Bs),[B|Bs1]) :-
	!,
	conj_to_list(Bs,Bs1).
conj_to_list(B,[B]).

orig_clauses(X) :- 
	end_of_file(X),
	!.
orig_clauses((A :- B)) :- 
	!,
	canonical((A :- B)),
	displayqRClauses([(A :- B)]),
	read(C1),
	!,
	orig_clauses(C1).
orig_clauses(A) :- 
        canonical((A :- true)),
        displayqRClauses([(A :- true)]),
	read(C1),
	!,
	orig_clauses(C1).
