% Optimised version of dg3.pl produced by SP.
% JPG: 17.9.93

% Algorithm to find the strongly connected components of a directed graph
% Taken from "Graph Algorithms" by Shimon Even, Pitman (1979), page 65.
%
% The algorithm has complexity k1.E + k2.V + k3 where E is number
% of edges and V is number of vertices (see R. Tarjan:  SIAM J. Computing;
% Vol 1. No 2. p. 146-160).
%
% (Note: the complexity of this version may be a bit worse because it uses
% some less efficient data structures).
%
% The graph is given by the dependency relation of a logic program.
% I.e. the vertices are predicate names and there is an edge p-q if 
% q occurs in the body of a clause with head p.

% JPG: 16.9.93

% strong_connected_components(Ps,Cs):  
%	Ps:  a list of vertices
%	Cs:  a list of components, each labelled as recursive or non-recursive

%:- mode strong_connected_components(i,o).
%:- mode init_depends_1(i,?).
%:- mode next_vertex_1(i,o).
%:- mode immed_depends_1(i,i,i,o).
%:- mode step3_1(i,o).
%:- mode step4_1(i,i,o).
%:- mode step5_1(i,o).
%:- mode step7_1(i,o).
%:- mode step8_1(i,o).
%:- mode step9_1(i,o).
%:- mode user_clauses_1(i,i,o).
%:- mode update_vertex_1(i,i,o).
%%:- mode get_vertex_1(i,i,?).
%:- mode body_preds_1(i,?,o,o).
%:- mode bodylits_1(i,?,o,o).
%:- mode rec_body_1(i,i,?).
%:- mode insertp_1(i,i,i,o).
%:- mode minimum_1(i,i,o).
%%:- mode newLV_1(i,i,o).
%:- mode memb1_1(i,i).
%%:- mode newstart_1(i,o).
%
strong_connected_components(nil,nil) :-
      !.
strong_connected_components(X0,X1) :-
      init_depends_1(X0,[X2 | X3]),
      next_vertex_1(state([X2 | X3],X2,nil,0),X4),
      step3_1(X4,X1),
      recursive_classify_1(X1).
init_depends_1([X0 | X1],[vertex(X0,X2,0,0,undef) | X4]) :-
      immed_depends_1(1,X0,nil,X2),
      init_depends_1(X1,X4).
init_depends_1(nil,nil) :-
      true.
next_vertex_1(state(X0,vertex(X1,X2,X3,X4,X5),X6,X7),
		state(X8,vertex(X1,X2,X9,X9,X5),[X1 | X6],X9)) :-
      X9 is X7 + 1,
      update_vertex_1([X1,X2,X9,X9,X5],X0,X8).
step3_1(state(X0,vertex(X1,nil,X2,X3,X4),X5,X6),X7) :-
      !,
      step7_1(state(X0,vertex(X1,nil,X2,X3,X4),X5,X6),X7).
step3_1(state(X0,vertex(X1,[X2 | X3],X4,X5,X6),X7,X8),X9) :-
      get_vertex_1(X2,X0,X10),
      update_vertex_1([X1,X3,X4,X5,X6],X0,X11),
      step4_1(state(X11,vertex(X1,X3,X4,X5,X6),X7,X8),X10,X9).
recursive_classify_1(nil) :-
      true.
recursive_classify_1([(recursive , [X0,X1 | X2]) | X3]) :-
      !,
      recursive_classify_1(X3).
recursive_classify_1([(recursive , [X0 / X1]) | X2]) :-
      functor(X3,X0,X1),
      ub_clause(X3,X4),
      rec_body_1(X0,X1,X4),
      !,
      recursive_classify_1(X2).
recursive_classify_1([(non_recursive , X0) | X1]) :-
      recursive_classify_1(X1).
immed_depends_1(0,X0,X1,X1) :-
      true.
immed_depends_1(X0,X1 / X2,X3,X4) :-
      X0 > 0,
      user_clauses_1(X1,X2,X5),
      body_preds_1(X0,X5,X3,X4).
update_vertex_1([X0,X1,X2,X3,X4],[vertex(X0,X5,X6,X7,X8) | X9],
		[vertex(X0,X1,X2,X3,X4) | X9]) :-
      !.
update_vertex_1(X1,[X5 | X6],[X5 | X7]) :-
      update_vertex_1(X1,X6,X7).
step7_1(state(X0,vertex(X1,X2,X3,X3,X4),X5,X6),[(X7 , X8) | X9]) :-
      !,
      pop_1(X5,X1,X10,X8),
      step8_1(state(X0,vertex(X1,X2,X3,X3,X4),X10,X6),X9).
step7_1(X0,X1) :-
      step8_1(X0,X1).
get_vertex_1(X0,[vertex(X0,X1,X2,X3,X4) | X5],vertex(X0,X1,X2,X3,X4)) :-
      !.
get_vertex_1(X0,[X1 | X2],X3) :-
      get_vertex_1(X0,X2,X3).
step4_1(state(X0,vertex(X1,X2,X3,X4,X5),X6,X7),vertex(X8,X9,0,X10,X11),X12) :-
      update_vertex_1([X8,X9,0,X10,X1],X0,X13),
      !,
      next_vertex_1(state(X13,vertex(X8,X9,0,X10,X1),X6,X7),X14),
      step3_1(X14,X12).
step4_1(X0,X1,X2) :-
      step5_1(X0,X1,X2).
rec_body_1(X0,X1,(X2 , X3)) :-
      functor(X2,X0,X1),
      !.
rec_body_1(X0,X1,(X2 , X3)) :-
      !,
      rec_body_1(X0,X1,X3).
rec_body_1(X0,X1,X2) :-
      functor(X2,X0,X1).
user_clauses_1(X0,X1,X2) :-
      functor(X3,X0,X1),
      findall((X3 :- X4),clause(X3,X4),X2),
      !.
user_clauses_1(X0,X1,nil) :-
      true.
body_preds_1(X0,[(X1 :- X2) | X3],X4,X5) :-
      bodylits_1(X0,X2,X4,X6),
      body_preds_1(X0,X3,X6,X5).
body_preds_1(X0,nil,X1,X1) :-
      true.
pop_1([X0 | X1],X0,X1,[X0]) :-
      !.
pop_1([X0 | X1],X2,X3,[X0 | X4]) :-
      pop_1(X1,X2,X3,X4).
step8_1(state(X0,vertex(X1,X2,X3,X4,X5),X6,X7),X8) :-
      X5 \== undef,
      !,
      get_vertex_1(X5,X0,vertex(X5,X9,X10,X11,X12)),
      minimum_1(X11,X4,X13),
      update_vertex_1([X5,X9,X10,X13,X12],X0,X14),
      step3_1(state(X14,vertex(X5,X9,X10,X13,X12),X6,X7),X8).
step8_1(X0,X1) :-
      step9_1(X0,X1).
step5_1(state(X0,vertex(X1,X2,X3,X4,X5),X6,X7),
		vertex(X8,X9,X10,X11,X12),X13) :-
      X10 > X3,
      !,
      step3_1(state(X0,vertex(X1,X2,X3,X4,X5),X6,X7),X13).
step5_1(state(X0,X1,X2,X3),vertex(X4,X5,X6,X7,X8),X9) :-
       \+ memb1_1(X4,X2),
      !,
      step3_1(state(X0,X1,X2,X3),X9).
step5_1(X0,X1,X2) :-
      newLV_1(X0,X1,X3),
      step3_1(X3,X2).
bodylits_1(X0,(X1 , X2),X3,X4) :-
       \+ sp_builtin(X1),
      !,
      functor(X1,X5,X6),
      insertp_1(X5,X6,X3,X7),
      X8 is X0 - 1,
      immed_depends_1(X8,X5 / X6,X7,X9),
      bodylits_1(X0,X2,X9,X4).
bodylits_1(X0,(X1 , X2),X3,X4) :-
      !,
      bodylits_1(X0,X2,X3,X4).
bodylits_1(X0,X1,X2,X3) :-
       \+ sp_builtin(X1),
      !,
      functor(X1,X4,X5),
      insertp_1(X4,X5,X2,X6),
      X7 is X0 - 1,
      immed_depends_1(X7,X4 / X5,X6,X3).
bodylits_1(X0,X1,X2,X2) :-
      true.
minimum_1(X0,X1,X0) :-
      X0 =< X1.
minimum_1(X0,X1,X1) :-
      X0 > X1.
step9_1(state(X0,X1,X2,X3),X4) :-
      newstart_1(X0,X5),
      !,
      next_vertex_1(state(X0,X5,X2,X3),X6),
      step3_1(X6,X4).
step9_1(X0,nil) :-
      true.
memb1_1(X0,[X0 | X1]) :-
      !.
memb1_1(X0,[X1 | X2]) :-
      memb1_1(X0,X2).
newLV_1(state(X0,vertex(X1,X2,X3,X4,X5),X6,X7),
		vertex(X8,X9,X10,X11,X12),
		state(X13,vertex(X1,X2,X3,X14,X5),X6,X7)) :-
      minimum_1(X4,X10,X14),
      update_vertex_1([X1,X2,X3,X14,X5],X0,X13).
insertp_1(X0,X1,X2,X2) :-
      memb1_1(X0 / X1,X2),
      !.
insertp_1(X0,X1,X2,[X0 / X1 | X2]) :-
      true.
newstart_1([vertex(X0,X1,0,X2,X3) | X4],vertex(X0,X1,0,X2,X3)) :-
      !.
newstart_1([X0 | X1],X2) :-
      newstart_1(X1,X2).

