/* intersect(T1,T2, P):  P is a procedure for a type which is the intersection
   of T1 and T2 */

/* JPG: 16.1.92 */

intersect(T1,T2,T1,[],D,N,N) :-
	subtype(T1,T2,D),
	!.
intersect(T1,T2,T2,[],D,N,N) :-
	subtype(T2,T1,D),
	!.
/*
intersect(any,T2,T2,[],D,N,N) :-
	!.
intersect(T1,any,T1,[],D,N,N) :-
	!.
*/
intersect(T1,T2,T,[proc(T/1,Def)|Subdefs],D,N,N3) :-
	def(T1/1,P1,D),
	def(T2/1,P2,D),
	newname(N,T),
	N1 is N+1,
	!,
	lower_type(P1,P2,Def,T,pairs([(T1,T2,T)|S],More),N1,N2,D),
%	(Def = [], !, write('empty def'),nl,write((T1,T2,D)),nl, trace; true),
	Def \== [],
	iterate_intersect(More,[(T1,T2,T)|S],Subdefs,N2,N3,D).

iterate_intersect([],_,[],N,N,_) :-
	!.
iterate_intersect(M,S,Defs,N,N2,D) :-
	each_intersect(M,Defs1, M1,S,N,N1,D),
	iterate_intersect(M1,S,Defs2,N1,N2,D),
	app(Defs1,Defs2,Defs).

each_intersect([],[],_,_,N,N,_) :-
	!.
each_intersect([(T1,T2,T)|Ps],[proc(T/1,Def)|Defs],More,S,N,N3,D) :-
	def(T1/1,P1,D),
	def(T2/1,P2,D),
	N1 is N+1,
	!,
	lower_type(P1,P2,Def,T,pairs(S,More),N1,N2,D),
	each_intersect(Ps,Defs,More,S,N2,N3,D).

lower_type([],_,[],_,_,N,N,_) :-
	!.
lower_type(_,[],[],_,_,N,N,_) :-
	!.
lower_type([(H :- true)|P1],P2,Newdef,T,More,N,N2,D) :-
        H =.. [_,K],
        basic_type(K),
        !,
	remove_any_type_instance(K,P1,P2,H,Newdef,T,More,N,N2,D).
lower_type([(H :- B1)|P1],P2,[(S :- B)|Def],T,More,N,N2,D) :-
	removeclause(H,B2,P2,P3),
	!,
	H =.. [_,K],
	S =.. [T,K],
	intersect_body(B1,B2,B,More,N,N1,D),
	lower_type(P1,P3,Def,T,More,N1,N2,D).
lower_type([_|P1],P2,Def,T,More,N,N2,D) :-
	lower_type(P1,P2,Def,T,More,N,N2,D).

remove_any_type_instance(K,P1,P2,H,[(S :- true)|Def],T,More,N,N2,D) :-
        remove_type_instance(K,P2,K1,P3),
	!,
        S =.. [T,K1],
        lower_type([(H :- true)|P1],P3,Def,T,More,N,N2,D).
remove_any_type_instance(_,P1,P2,_,Def,T,More,N,N2,D) :-
        lower_type(P1,P2,Def,T,More,N,N2,D).

remove_type_instance(K,[(H :- true)|P],K1,P) :-
        H =.. [_,K1],
        type_check(K1,K),
        !.
remove_type_instance(K,[C|P],K1,[C|P1]) :-
        remove_type_instance(K,P,K1,P1).


intersect_body(true,true,true,_,N,N,_) :-
	!.
intersect_body((B,Bs),(C,Cs),(D,Ds),More,N,N2,P) :-
	!,
	intersect_atom(B,C,D,More,N,N1,P),
	intersect_body(Bs,Cs,Ds,More,N1,N2,P).
intersect_body(B,C,D,More,N,N1,P) :-
	intersect_atom(B,C,D,More,N,N1,P).

intersect_atom(B,C,D,More,N,N1,P) :-
        B =.. [R,X],
        C =.. [S,X],
        next_intersect(R,S,T,More,N,N1,P),
        D =.. [T,X].

next_intersect(R,S,R,_,N,N,D) :-
        subtype(R,S,D),
        !.
next_intersect(R,S,S,_,N,N,D) :-
        subtype(S,R,D),
        !.
next_intersect(R,S,T,More,N,N1,_) :-
        get_name(R,S,T,More,N,N1).


