/* regularise(P,D):  D is set of unary clauses.
   P is set of RUL clauses. */

/* JPG: 18.8.93 adapted from ub_merge.pl */ 


is_regularise(R,D) :-
	allclauses(D,Cls),
	regularise_clauses(Cls,R1,S,M),
	iterate_is_merge(M,S,R1,R).

iterate_is_merge([],_,R,R) :-
	!.
iterate_is_merge(M,S,R,R2) :-
	make_new_clauses(M,R,R1),
	regularise_clauses(R1,R2,S,M1),
	app(R,R2,R3),
	iterate_is_merge(M1,S,R3,R).

regularise_clauses([],[],_,_).
regularise_clauses([(H :- B)|Cls],R,S,M) :-
	repeat_body_var(B,X,B1,Preds),
	!,
	new_is_name(Preds,S,M),
	regularise_clauses([(H :- B1)|Cls],R,S,M) :-
regularise_clauses([(H :- B)|Cls],[(H :- B)|R],S,M) :-
	regularise_clauses(Cls,R,S,M).
	
make_new_clauses([(Ps,T)|M],R,R1) :-
	
each_is_merge([],[],_,__) :-
	!.
each_is_merge([(T,T,T)|Ps],[proc(T/1,Def)|Defs],More,S,D) :-
	!,
	rul_defs(T/1,P1,D),
	is_merge_rul(P1,Def,pairs(S,More),D),
	each_is_merge(Ps,Defs,More,S,N1,N2,D).
each_is_merge([(T1,T2,T3)|Ps],[proc(T3/1,Def)|Defs],More,S,D) :-
	!,
	rul_defs(T1/1,P1,D),
	rul_defs(T2/1,P2,D),
	app(P1,P2,P3),
	renametypes(P3,T3,P4),
	is_merge_rul(P4,Def,pairs(S,More),D),
	each_is_merge(Ps,Defs,More,S,N1,N2,D).

is_merge_rul(P,Q,More,D) :-
	remove((H :- B1),P,P1),
	removeclause(H,B1,P1,P2),	% two identical clauses 
	!,
	is_merge_rul([(H :- B1)|P2],Q,More,D).
is_merge_rul(P,Q,More,D) :-
	remove((H :- B1),P,P1),
	removeclause(H,B2,P1,P2),	% two clauses with same head 
	!,
	is_body_merge(B1,B2,B,More,D),
	is_merge_rul([(H :- B)|P2],Q,More,N1,N2,D).
is_merge_rul(P,P,__).

is_body_merge(true,true,true,__) :-
	!.
is_body_merge((B,Bs),(C,Cs),(D,Ds),More,P) :-
	!,
	is_atom_merge(B,C,D,More,P),
	is_body_merge(Bs,Cs,Ds,More,N1,N2,P).
is_body_merge(B,C,D,More,P) :-
	is_atom_merge(B,C,D,More,P).

is_atom_merge(_,C,C,__) :-
	C =.. [any,_],
	!.
is_atom_merge(B,_,B,__) :-
	B =.. [any,_],
	!.
is_atom_merge(C,C,C,S,_) :-
	!,
	insert_pair((C,C,C),S).
is_atom_merge(B,C,D,More,_) :-
	B =.. [R,X],
	C =.. [S,X],
	get_name(R,S,T,More),
	D =.. [T,X].

all_proc_names([proc(T/1,_)|Ps],S,M) :-
	memb1((T,T,T),S),
	memb1((T,T,T),M),
        all_proc_names(Ps,S,M).
all_proc_names([Qs|Ps],S,M) :-
        all_proc_names(Qs,S,M),
        all_proc_names(Ps,S,M).
all_proc_names([],_,_).

rul_defs(T/N,P,[proc(T/N,P1)|Ps]) :-
	!,
        rul_defs(T/N,P2,Ps),
	app(P1,P2,P).
rul_defs(T,P,[proc(_,_)|Ps]) :-
	!,
        rul_defs(T,P,Ps).
rul_defs(T,P,[Qs|Ps]) :-
        rul_defs(T,P1,Qs),
        rul_defs(T,P2,Ps),
	app(P1,P2,P).
rul_defs(_,[],[]).

renametypes([],_,[]) :-
	!.
renametypes([(H :- B)|Ps],T,[(S :- B)|Qs]) :-
	H =.. [_,X],
	S =.. [T,X],
	renametypes(Ps,T,Qs).

get_name(R,S,T,More) :-
        R @=< S,
        !,
        insert_pair((R,S,T),More).
get_name(R,S,T,More) :-
        insert_pair((S,R,T),More).


insert_pair(P,pairs(S,M)) :-
        insert_pair1(P,S,M).

insert_pair1((A,A,A),S,M :-
        var(S),
        !,
        S = [(A,A,A)|_],
        memb1((A,A,A),M).
insert_pair1((A,B,C),S,M) :-
        var(S),
        !,
        newname(N,C),
        N1 is N+1,
        S = [(A,B,C)|_],
        memb1((A,B,C),M).
insert_pair1(P,[P|_],_ :-
        !.
insert_pair1(P,[_|S],M) :-
        insert_pair1(P,S,M).



