/* subtype(T1,T2,D):  T1 is a sub-type of T2 with set of type defs D */
/* T1 and T2 are atoms whose procedures are
   defined by Regular Unary Logic Programs */

/* JPG:  15.1.92   */

%:- setdebug.
% :- dynamic user_proc/3.

eqtype(T1,T2,D) :-
	subtype(T1,T2,D),
	subtype(T2,T1,D).

subtype(T1,T2,D) :-
	subtype(T1,T2,[],_,D).

subtype(_,any,S,S,_) :-
	!.
subtype(T,T,S,S,_) :-
	!.
subtype(T1,T2,S,S,_) :-
	loopsubtype(T1,T2,S),
	!.
subtype(T1,T2,S,S1,D) :-
	T1 \== any,
	T2 \== any,
	def(T1/1,P1,D),
	def(T2/1,P2,D),
	!,
	subtype1(P1,P2,[(T1,T2)|S],S1,D).


subtype1([],_,S,S,_).
subtype1([(H :- B)|P1],P2,S,S2,D) :-
	removeclause(H,B1,P2,P3),
	subtype_pairs(B,B1,S,S1,D),
	subtype1(P1,P3,S1,S2,D).
	
subtype_pairs(true,true,S,S,_) :-
	!.
subtype_pairs((B,Bs),(D,Ds),S,S2,P) :-
	!,
	subtype2(B,D,S,S1,P),
	subtype_pairs(Bs,Ds,S1,S2,P).
subtype_pairs(B,D,S,S1,P) :-
	subtype2(B,D,S,S1,P).

subtype2(B,D,S,S1,P) :-
	functor(B,T1,1),
	functor(D,T2,1),
	subtype(T1,T2,S,S1,P).

removeclause(H,B,[(G :- B)|P],P) :-
	H =.. [_,T1],
	G =.. [_,T2],
	functor(T1,F,N),
	functor(T2,F,N),
	!.
removeclause(H,true,[(G :- true)|P],[(G :- true)|P]) :-
	G =.. [_,T2],
	basic_type(T2),
	H =.. [_,T1],
	type_check(T1,T2),
	!.
removeclause(H,B,[X|P],[X|P1]) :-
	removeclause(H,B,P,P1).

loopsubtype(X,Y,[(X,Y)|_]) :-
	!.
loopsubtype(X,Y,[_|S]) :-
	loopsubtype(X,Y,S).

user_clauses(P/N,Cls) :-
        functor(H,P,N),
        findall((H:-B),ub_clause(H,B),Cls),
	!.
user_clauses(_,[]).


	

def(numexpr/1,P,_) :-
	!,
	findall((numexpr(H) :- B),clause(numexpr(H),B),P).
def(numeric/1,P,_) :-
	!,
	findall((numeric(H) :- B),clause(numeric(H),B),P).
def(T,P,D) :-
	def1(T,P,D),
	!.
def(T,P,_) :-
	user_clauses(T,P),
	canonical(P),
	P \== [],
	!.
def(T,_,_) :-
	write(T),write(' not found'),nl,fail.

/*
def1(T,P,[[]|Ps]) :-
	def1(T,P,Ps),
	!.
def1(T,P,[[Q|Qs]|_]) :-
	def1(T,P,[Q|Qs]),
	!.
def1(T,P,[[_|_]|Ps]) :-
	def1(T,P,Ps),
	!.
*/

def1(T/1,P,[proc(T/1,P)|_]) :-
	!.
def1(T,P,[proc(_,_)|D]) :-
	def1(T,P,D).

	
	
basic_type(numeric).

:- dynamic numexpr/1.
:- dynamic numeric/1.

numeric(numeric) :- true.

numexpr(numeric) :- true.
numexpr('X0'+'X1') :- numexpr('X0'),numexpr('X1').
numexpr('X0'-'X1') :- numexpr('X0'),numexpr('X1').
numexpr('X0'/'X1') :- numexpr('X0'),numexpr('X1').
numexpr('X0'*'X1') :- numexpr('X0'),numexpr('X1').
numexpr('X0' mod 'X1') :- numexpr('X0'),numexpr('X1').
numexpr(-'X0') :- numexpr('X0').



type_check(X,numeric) :-
	number(X).


/* experimental treatment of any, alpha_j etc.  

basic_type(any).

type_check(X,any) :-
	functor(X,_,N),
	N > 0.
type_check(X,any) :-
	functor(X,F,0),
	\+ alpha_const(F).

alpha_const(X) :-
	name(X,Xn),
	name(alpha,A),
	app(A,_,Xn).

*/
