% JPG - 14.2.92
% Compute a regular approximation of a given program.
% For a description of the algorithm,
% see Gallagher and de Waal:  "Regular Approximations of
% Logic Programs and Their Uses";  CSTR-92-06.

%:- setdebug.
 
tp_fix(P) :-
	nl,
	tp_fix(P,F),
	display_fix(F).

tp_fix(Ps,R) :-
	init(Ps,R0),
	tp_iterate([[]],Ps,R0,R,10,_).

init([_|Ps],[[]|R]) :-
	init(Ps,R).
init([],[]).

tp_iterate([],_,R,R,K,K) :-
	!.
tp_iterate(Nofix,Ps,R1,R,K,K3) :-
	writestate(R1),
	tp_alpha(Ps,Nofix,R1,R1,R2,Chs,K,K1),
	fold(R2,R3,Chs,K1,K2),
	countdefs(R3),
	check_fix(Ps,R1,R3,Fix),
	!,
	tp_iterate(Fix,Ps,R3,R,K2,K3).

countdefs([R|Rs]) :-
	sp_length(R,N),write(N),write(' '),
	!,
	countdefs(Rs).
countdefs([]) :-
	nl.

fold([],[],[],K,K).
fold([RP|R1],[RP|R3],[nochange|Chs],K,K2) :-
	!,
	fold(R1,R3,Chs,K,K2).
fold([RP|R1],[RP1|R3],[change|Chs],K,K2) :-
	foldp(RP,RP1,K,K1),
	fold(R1,R3,Chs,K1,K2).
	
:- dynamic level/1.

level(1).

foldp(RP,RP3,K,K2) :-
	depends(T1,T2,TPs,RP),
	level(Level),
	same_functor_structure(Level,T2,T1,TPs),
	recursive(T1,T2,TPs,RP,RP1,K,K1),
	!,
	toptype(RP1,T),
	rdefs(T,RP2,RP1),
	foldp(RP2,RP3,K1,K2).
foldp(R,R,K,K).

recursive(T1,T2,TPs,RP,RP2,K,K) :-
	subtype(T2,T1,TPs),
	!,
	recursive1(T1,T2,TPs,RP,RP2,K,K).
recursive(T1,T2,TPs,RP,RP2,K,K1) :-
        ub1(T1,T2,T,D,TPs,K,K1),
	app(RP,D,RP1),
	normalise_defs(RP1,T1,T,RP2).

recursive1(T1,T2,TPs,RP,RP2,K,K) :-
	subtype(T1,T2,TPs),
	!,
	normalise_defs(RP,T2,T1,RP2).
recursive1(T1,T2,TPs,RP,RP2,K,K) :-
	normalise_subdefs(TPs,T2,T1,TPs1),
	replace_procs(TPs1,RP,RP2).
	
normalise_subdefs(TPs,T2,T1,TPs1) :-
	depend_path(TPs,T1,[T2,T3|_]),
	!,
	def(T3/1,Cls,TPs),
	normalise(Cls,T2,T1,Cls1),
	replace_proc(proc(T3/1,Cls1),TPs,TPs2),
	!,
	normalise_subdefs(TPs2,T2,T1,TPs1).
normalise_subdefs(TPs,_,_,TPs).
	
depend_path([proc(T1/1,Cls)|Ps],T1,Path) :-
	body_pred(Cls,T),
	T1 \== T,
	\+ anypred(T),
	dep_path_cls(Ps,T,Path,[T1]).

dep_path_cls(_,T,[T|D],D).
dep_path_cls(TPs,T,Path,D) :-
	def(T/1,Cls,TPs),
	body_pred(Cls,T1),
	\+ memb2(T1,[T|D]),
	\+ anypred(T1),
	dep_path_cls(TPs,T1,Path,[T|D]).
	
body_pred([(_ :- B)|_],T) :-
        bodytype(B,T).
body_pred([_|Cs],T) :-
        body_pred(Cs,T).

bodytype((B,_),T) :-
        functor(B,T,1).
bodytype((_,Bs),T) :-
	!,
        bodytype(Bs,T).
bodytype(B,T) :-
        functor(B,T,1).


	
replace_procs([],R,R).
replace_procs([P|Ps],R,R1) :-
	replace_proc(P,R,R2),
	replace_procs(Ps,R2,R1).

replace_proc(proc(T1/1,Cls),[proc(T1/1,_)|R],[proc(T1/1,Cls)|R]) :-
	!.
replace_proc(P,[P1|R],[P1|R1]) :-
	replace_proc(P,R,R1).
replace_proc(P,[],[P]).


less_functors(T1,T2,R) :-
	\+ morefunctors(T1,T2,R).

morefunctors(T1,T2,R) :-
	headfunctor(F,T1,R),
	\+ headfunctor(F,T2,R).

same_functor_structure(1,T1,T2,RP) :-
	\+ morefunctors(T1,T2,RP),
	\+ morefunctors(T2,T1,RP).

same_functor_structure(2,T1,T2,R) :-
	\+ diffstructure(T1,T2,R),
	\+ morefunctors(T2,T1,R).

diffstructure(T1,T2,R) :-
	typeclause(T1/1,R,(H :- B)),
	H =.. [T1,X],
	functor(X,F,N),
	\+ corresponding_clause(T2,F/N,B,R).

corresponding_clause(T,F/N,B,R) :-
	typeclause(T/1,R,(H :- B1)),
	H =.. [T,X],
	functor(X,F,N),
	same_body_structure(B,B1,R).

same_body_structure(true,true,_).
same_body_structure((A1,Bs),(A2,Bs1),R) :-
	anyatom(A1,T),
	anyatom(A2,T),
	!,
	same_body_structure(Bs,Bs1,R).
same_body_structure((B,Bs),(B1,Bs1),R) :-
	!,
	functor(B,T,1),
	functor(B1,T1,1),
	\+ anypred(T),
	\+ anypred(T1),
	same_functor_structure(1,T,T1,R),
	same_body_structure(Bs,Bs1,R).
same_body_structure(B,B1,R) :-
	functor(B,T,1),
	functor(B1,T1,1),
	\+ anypred(T),
	\+ anypred(T1),
	same_functor_structure(1,T,T1,R).

headfunctor(F/N,T,Defs) :-
        typeclause(T/1,Defs,(H :- _)),
        H =.. [T,X],
        functor(X,F,N).


	
depends(T1,T2,R1,[_|P]) :-
	nested_memb(proc(T1/1,_),P),
	iterate_defs([T1],[T1|S],R1,[],P),
	makelist(S),
	memb2(T2,S).

makelist([]) :-
	!.
makelist([_|X]) :-
	makelist(X).

body_t([(_ :- B)|_],T3) :-
	lit(A,B),
	functor(A,T3,1).
body_t([_|Cls],T3) :-
	body_t(Cls,T3).

	
tp_alpha([],_,_,[],[],[],K,K).
tp_alpha([P|Ps],Nofix,R,[PR|PRs],[UP|R1],[Ch|Chs],K,K3) :-
	user_clauses(P,Cls),
	infer_proc(Cls,P,1,Nofix,R,As,K,K1),
	ub_proc(As,[As|R],PR,UP,Ch,K1,K2),
	tp_alpha(Ps,Nofix,R,PRs,R1,Chs,K2,K3).

infer_proc([(_ :- B)|Cls],J,Nofix,R,As,K,K1) :-	% if no predicate in the body
	nochange(B,Nofix),			% changed, skip clause
	!,
	J1 is J+1,
	infer_proc(Cls,J1,Nofix,R,As,K,K1).
infer_proc([(H :- B)|Cls],J,Nofix,R,[A1|As],K,K3) :-
	tsolve(B,R,T),
	unfold_body(T,R,T1),
	intersect_body(T1,R,T2,S,K,K1),
	!,
	headtype(H,J,T2,_,A,K1,K2), 	
	app(A,S,A1),
	J1 is J+1,
	infer_proc(Cls,J1,Nofix,R,As,K2,K3).
infer_proc([_|Cls],J,Nofix,R,As,K,K1) :-		% if no solution in R.
	J1 is J+1,
	infer_proc(Cls,J1,Nofix,R,As,K,K1).
infer_proc([],_,_,_,[],K,K).

nochange(true,[[]]) :-  	% unit clauses handled on 1st iteration
	!,
	fail.
nochange(true,_) :-
	!.
nochange(B,Nofix) :-
	preds(B,Ps),
	\+ commonpred(Ps,Nofix).

preds(true,[]) :-
	!.
preds((B,Bs), [P/N|Ps]) :-
	!,
	functor(B,P,N),
        preds(Bs,Ps).
preds(B,[P/N]) :-
	functor(B,P,N).

commonpred(X,Y) :-
	memb2(Z,X),
	memb2(Z,Y).
	
tsolve(true,_,true) :-
	!.
tsolve((B,Bs),R,T) :-
	!,
	tsolve(B,R,T1),
	tsolve(Bs,R,T2),
	joingoals(T1,T2,T).
tsolve(B,_,Bs) :-
        sp_builtin(B),
        !,
        any_term(B,Bs).
tsolve(B,R,T1) :-
	functor(B,S,N),
	headclause(S/N,R,(H :- T)),
	melt((H :- T),(H1 :- T1)),
	H1 =.. [_,B].

unfold_body(true,_,true) :-
	!.
unfold_body((B,Bs),R,C) :-
	!,
	unfold_body(B,R,C1),
	unfold_body(Bs,R,C2),
	joingoals(C1,C2,C).
unfold_body(B,_,B) :-
	B =.. [_,X],
	var(X),
	!.
unfold_body(B,_,Bs) :-
	anyatom(B,A),
	!,
	any_term1(A,B,Bs).
unfold_body(B,R,B2) :-
	B =.. [T,_],
	typeclause(T/1,R,(H :- B1)),
	melt((H :- B1),(B :- B3)),
	unfold_body(B3,R,B2).

any_term(B,Bs) :-
	vars(B,Vs),
	any_vars(Vs,Bs).

any_vars([],true) :-
	!.
any_vars([X],any(X)) :-
	!.
any_vars([X|Xs],(any(X),Bs)) :-
	any_vars(Xs,Bs).
	
any_term1(A,B,Bs) :-
	B =.. [F|Args],
	name(A,An),
	name(F,Fn),
	app(An,[95|Fn],S),
	any_term_path(Args,1,Bs,S).

any_term_path([],_,true,_) :-
	!.
any_term_path([X|Xs],M,Bs,S) :-
	any_arg_path(X,M,B1,S),
	M1 is M+1,
	any_term_path(Xs,M1,B2,S),
	joingoals(B1,B2,Bs).

any_arg_path(X,M,A,S) :-
	var(X),
	!,
	name(M,Mn),
	app(S,[95|Mn],S1),
	name(T,S1),
	A =.. [T,X].
any_arg_path(X,M,Bs,S) :-
	X =.. [F|Args],
        name(F,Fn),
        app(S,[95|Fn],S1),
        any_term_path(Args,1,Bs,S1).

intersect_body(true,_,true,[],K,K) :-
	!.
intersect_body((B,Bs),R,C,I,K,K2) :-
	reoccurs(B,Bs,B1,Bs1),
	!,
	B =.. [T1,X],
	B1 =.. [T2,X],
	intersect(T1,T2,T3,P,R,K,K1),
	B2 =.. [T3,X],
        intersect_body((B2,Bs1),[P|R],C,J,K1,K2),
	app(P,J,I).
intersect_body((B,Bs),R,(B,Bs1),I,K,K1) :-
	!,
	intersect_body(Bs,R,Bs1,I,K,K1).
intersect_body(B,_,B,[],K,K).
	
	
reoccurs(B,(B1,Bs),B1,Bs) :-
	B =.. [_,X],
	B1 =.. [_,Y],
	X == Y,
	!.
reoccurs(B,(B2,Bs),B1,(B2,Bs1)) :-
	!,
	reoccurs(B,Bs,B1,Bs1).
reoccurs(B,B1,B1,true) :-
	B =.. [_,X],
	B1 =.. [_,Y],
	X == Y,
	!.
	
headtype(T,J, Body,P,[proc(P/1,[(S :- B)])|Argtypes],K,K2) :- 
	functor(T,F,N),
	functor(X,F,N),
	newname(K,P),
	S =.. [P,X],
	canonical(S),
	K1 is K+1,
	copyterm(T,TC),
	canonical(TC),
	argtypes(T,TC,F/N,J,Body,X,B,1,N,K1,K2,Argtypes).
	
termtype(T,TC,J,Body,[proc(P/1,[(S :- B)])|Argtypes],K,K2) :- 
	functor(T,F,N),
	functor(X,F,N),
	newname(K,P),
	S =.. [P,X],
	canonical(S),
	K1 is K+1,
	argtypes(T,TC,Body,X,B,1,N,K1,K2,Argtypes).

argtypes(_,_,_,_,_,_,true,_,0,K,K,[]).
argtypes(T,TC,F,J,Body,X,B,N,N,K,K1,Argtypes) :-
	argtype(T,TC,F,J,Body,X,B,N,K,K1,Argtypes).
argtypes(T,TC,F,J,Body,X,(B,Bs),M,N,K,K2,Argtypes) :-
	M < N,
	argtype(T,TC,F,J,Body,X,B,M,K,K1,Argtype),
	M1 is M+1,
	argtypes(T,TC,F,J,Body,X,Bs,M1,N,K1,K2,Argtypes1),
	app(Argtype,Argtypes1,Argtypes). 
	
argtype(T,TC,F,J,Body,X,B,M,K,K,[]) :-
	arg(M,T,Xm),
	arg(M,TC,Am),
	var(Xm),
	!,
	arg(M,X,Ym),
	giventype(Xm,Am,F,J,Body,R),
	B =.. [R,Ym].
argtype(T,TC,F,J,Body,X,B,M,K,K1,Types) :-
	arg(M,T,Xm),
	arg(M,T,Am),
	arg(M,X,Ym),
	termtype(Xm,Am,F,J,Body,Types,K,K1),
	newname(K,P),
	B =.. [P,Ym].

giventype(X,_,_,_,(B,_),T) :-
	B =.. [T,Y],
	X == Y,
	!.
giventype(X,A,F,J,(_,Bs),T) :-
	!,
	giventype(X,A,F,J,Bs,T).
giventype(X,_,_,_,B,T) :-
	B =.. [T,Y],
	X == Y,
	!.
giventype(_,A,F/N,J,_,Any) :- % should return type name for that variable
	name(A,An),
	name(F,Fn),
	name(N,Nn),
	name(any,Anyn),
	app(N,[95|An],S1),
	app(Fn,[95|S1],S2),
	app(Anyn,[95|S2],S3),
	name(Any,S3).
	
	
newname(K,P) :-
	name(K,Kn),
	name(t,T),
	app(T,Kn,Tn),
	name(P,Tn).


check_fix([],[],[],[]).
check_fix([_|Ps],[RP|R],[RP1|S],Qs) :-
        fixpoint(RP,RP1),
	!,
        check_fix(Ps,R,S,Qs).
check_fix([P|Ps],[_|R],[_|S],[P|Qs]) :-
        check_fix(Ps,R,S,Qs).

fixpoint([],[]) :-
	!.
fixpoint(R1,R2) :-
	toptype(R1,T1),
	toptype(R2,T2),
	subtype(T2,T1,[R1|R2]).
	
toptype([proc(T1/1,_)|_],T1).

ub_proc([],_,R1,R1,nochange,K,K) :-  
        !.
ub_proc([P],R,[proc(TP/1,_)|_],D,nochange,K,K) :-  % ub with current def
	toptype(P,Q),
	subtype(Q,TP,R),	% no increase over previous def
	!,
	rdefs(TP,D,R).
ub_proc([P],R,[proc(TP/1,_)|_],D,change,K,K1) :-  % upper bound with current def
	!,
	toptype(P,Q),
	ub1(TP,Q,Q1,D1,R,K,K1),
	rdefs(Q1,D,[D1|R]).
ub_proc([P],R,_,D,change,K,K) :-  
	!,
	toptype(P,Q),
	rdefs(Q,D,R).
ub_proc([P1,P2|Ps],R,R1,U,Ch,K,K2) :-
	toptype(P1,T1),
	toptype(P2,T2),
	subtype(T1,T2,R),
	!,
	ub_proc([P2|Ps],[P2|R],R1,U,Ch,K,K2).
ub_proc([P1,P2|Ps],R,R1,U,Ch,K,K2) :-
	toptype(P1,T1),
	toptype(P2,T2),
	subtype(T2,T1,R),
	!,
	ub_proc([P1|Ps],[P1|R],R1,U,Ch,K,K2).
ub_proc([P1,P2|Ps],R,R1,U,Ch,K,K2) :-
	toptype(P1,T1),
	toptype(P2,T2),
        ub1(T1,T2,_,D,R,K,K1),
	ub_proc([D|Ps],[D|R],R1,U,Ch,K1,K2).

headclause(P/N,[[proc(T/1,[(H :- B)])|_]|_],(H :- B)) :-
	H =.. [T,X],
	functor(X,P,N),
	!.
headclause(P,[_|Ps],Cl) :-
	headclause(P,Ps,Cl).


nested_memb(X,[X|_]).
nested_memb(X,[Y|_]) :-
	nested_memb(X,Y).
nested_memb(X,[_|Z]) :-
	nested_memb(X,Z).


typeclause(T/N,P,Cl) :-
	def(T/N,Q,P),
	!,
	memb2(Cl,Q).
	


merge_procs([],R,R).
merge_procs([P|Ps],R1,R2) :-
	insert_proc(P,R1,R3),
	merge_procs(Ps,R3,R2).

insert_proc(P,R,R) :-
	procmemb(P,R),
	!.
insert_proc(P,R,[P|R]).

procmemb(proc(F/N,_),[proc(F/N,_)|_]).
procmemb(P,[_|Ps]) :-
	procmemb(P,Ps).

:- dynamic traceon/0.

writestate(R) :-
	traceon,
	write(R),
	read(_),
	!,
	nl.
writestate(_) :-
	write('*'),
	sp_ttyflush.

lit(B, (B,_)).
lit(B, (_,Bs)) :-
        !,
        lit(B,Bs).
lit(B,B).

display_fix(Cls) :-
        nl,write('Enter filename: '),
        read(F),
        reg_filewrite(F,Cls).

reg_filewrite(tty,Cls) :-
        !,
	nl,write('Regular Approximation Program:'),nl,nl,
        displayqTypes(Cls),
	nl,nl.
reg_filewrite(F,Cls) :-
        tell(F),
        displayqTypes(Cls),
        told.

/* write the program to current output file  */

displayqTypes([[]|Ps]) :-
	displayqTypes(Ps).
displayqTypes([[Q|Qs]|Ps]) :-
	displayqPreddef(Q),
	displayqTypes(Qs),
	displayqTypes(Ps).
displayqTypes([proc(_,Qs)|Ps]) :-
	displayqRClauses(Qs),
	displayqTypes(Ps).
displayqTypes([]).

displayqPreddef(proc(_,[(H :- B)])) :-
	H =.. [_,A],
	displayqRClauses([(A :- B)]).


anypred(P) :-
	name(any,[A1,A2,A3]),
	name(P,[A1,A2,A3|_]).

anyatom(B,F) :-
	functor(B,F,1),
	anypred(F).
