/* regularise(P,D):  D is set of RUL clauses which 
   is not regular program.  P is RUL program. */

/* JPG: 16.1.92 */ 
/* JPG: 13.8.93 adapted from ub.pl */ 


regularise(P,Q/N,R0,D) :-
	name(Q,Qn),
	argtypenames(Qn,N,1,Ns),
	make_s_list([P|Ns],S,M),
	iterate_ub_merge(M,S,R,[],D),
	%iterate_ub_merge([(P,P,P)|_],[(P,P,P)|_],R,[],D),
	reverse_procs(R,R0,[]).

reverse_procs([],Y,Y).
reverse_procs([X|Xs],Y,Z) :-
	reverse_procs(Xs,Y,[X|Z]).

make_s_list([],_,_).
make_s_list([P|Ps],[(P,P,P)|Qs],[(P,P,P)|Rs]) :-
	make_s_list(Ps,Qs,Rs).

iterate_ub_merge([],_,R,R,_) :-
	!.
iterate_ub_merge(M,S,Defs,R,D) :-
	each_ub_merge(M,M1,S,D,R,Defs1),
	iterate_ub_merge(M1,S,Defs,Defs1,D).

each_ub_merge([],_,_,_,R,R) :-
	!.
each_ub_merge([(T,T,T)|Ps],More,S,D,R1,R) :-
	!,
	rul_defs(T/1,P1,D,R1),
	merge_rul(P1,Def,pairs(S,More)),
	each_ub_merge(Ps,More,S,D,[proc(T/1,Def)|R1],R).
each_ub_merge([(T1,T2,T3)|Ps],More,S,D,R1,R) :-
	!,
	rul_defs(T1/1,P1,D,R1),
	rul_defs(T2/1,P2,D,R1),
	app(P1,P2,P3),
	renametypes(P3,T3,P4),
	merge_rul(P4,Def,pairs(S,More)),
	each_ub_merge(Ps,More,S,D,[proc(T3/1,Def)|R1],R).

merge_rul(P,Q,More) :-
	remove((H :- B1),P,P1),
	removeclause(H,B1,P1,P2),	% two identical clauses 
	!,
	merge_rul([(H :- B1)|P2],Q,More).
merge_rul(P,Q,More) :-
	remove((H :- B1),P,P1),
	removeclause(H,B2,P1,P2),	% two clauses with same head 
	!,
	ub_body_merge(B1,B2,B,More),
	merge_rul([(H :- B)|P2],Q,More).
merge_rul(P,P,_).

ub_body_merge(true,true,true,_) :-
	!.
ub_body_merge((B,Bs),(C,Cs),(D,Ds),More) :-
	!,
	ub_atom_merge(B,C,D,More),
	ub_body_merge(Bs,Cs,Ds,More).
ub_body_merge(B,C,D,More) :-
	ub_atom_merge(B,C,D,More).

ub_atom_merge(_,C,C,_) :-
	C =.. [any,_],
	!.
ub_atom_merge(B,_,B,_) :-
	B =.. [any,_],
	!.
ub_atom_merge(C,C,C,S) :-
	C =.. [T,_],
	!,
	insert_ub_pair((T,T,T),S).
ub_atom_merge(B,C,D,More) :-
	B =.. [R,X],
	C =.. [S,X],
	get_ub_name(R,S,T,More),
	D =.. [T,X].

all_proc_names([proc(T/1,_)|Ps],S,M) :-
	memb1((T,T,T),S),
	memb1((T,T,T),M),
        all_proc_names(Ps,S,M).
all_proc_names([Qs|Ps],S,M) :-
        all_proc_names(Qs,S,M),
        all_proc_names(Ps,S,M).
all_proc_names([],_,_).

rul_defs(T/N,P,_,R) :-
	def1(T/N,P,R),
	!.
rul_defs(T/N,P,D,_) :-
	rul_defs(T/N,P,D).

rul_defs(T/N,P,[proc(T/N,P1)|Ps]) :-
	!,
        rul_defs(T/N,P2,Ps),
	app(P1,P2,P).
rul_defs(T,P,[proc(_,_)|Ps]) :-
	!,
        rul_defs(T,P,Ps).
rul_defs(T,P,[Qs|Ps]) :-
        rul_defs(T,P1,Qs),
        rul_defs(T,P2,Ps),
	app(P1,P2,P).
rul_defs(_,[],[]).

/*
rul_defs(T/N,P,[proc(T/N,P1)|Ps],Qs) :-
	!,
        rul_defs(T/N,P2,Ps,Qs),
	app(P1,P2,P).
rul_defs(T,P,[proc(A,B)|Ps],[proc(A,B)|Qs]) :-
	!,
        rul_defs(T,P,Ps,Qs).
rul_defs(T,P,[Qs|Ps],[Rs|Ss]) :-
        rul_defs(T,P1,Qs,Rs),
        rul_defs(T,P2,Ps,Ss),
	app(P1,P2,P).
rul_defs(_,[],[],[]).
*/

renametypes([],_,[]) :-
	!.
renametypes([(H :- B)|Ps],T,[(S :- B)|Qs]) :-
	H =.. [_,X],
	S =.. [T,X],
	renametypes(Ps,T,Qs).

get_ub_name(R,S,T,More) :-
        R @=< S,
        !,
        insert_ub_pair((R,S,T),More).
get_ub_name(R,S,T,More) :-
        insert_ub_pair((S,R,T),More).


insert_ub_pair(P,pairs(S,M)) :-
        insert_ub_pair1(P,S,M).

insert_ub_pair1((A,A,A),S,M) :-
        var(S),
        !,
        S = [(A,A,A)|_],
        memb1((A,A,A),M).
insert_ub_pair1((A,B,C),S,M) :-
        var(S),
        !,
        ub_name(A,B,C),
        S = [(A,B,C)|_],
        memb1((A,B,C),M).
insert_ub_pair1(P,[P|_],_) :-
        !.
insert_ub_pair1(P,[_|S],M) :-
        insert_ub_pair1(P,S,M).

ub_name(T1,T2,T) :-
	decompose_name(T1,D1),
        decompose_name(T2,D2),
	u_append(D1,D2,D3),
	u_normalise(D3,D4),
	compound_is_name(D4,C),
        name(T,C).

insertT(any,_,_) :-
        !.
insertT(T,S,M) :-
        var(S),
        !,
        memb1(T,M),
        S = [T|_].
insertT(T,[T|_],_) :-
        !.
insertT(T,[_|S],M) :-
        insertT(T,S,M).

/* get all clauses on which a predicate depends */

rdefs(T,Defs,P) :-
        iterate_defs([T],[T|_],Defs,[],P).

iterate_defs([],_,D,D,_) :-
        !.
iterate_defs(Ts,Ss,Defs,Defs2,P) :-
        each_def(Ts,Defs,Defs1,Ss,More,P),
        iterate_defs(More,Ss,Defs1,Defs2,P).

each_def([],D,D,_,_,_) :-
        !.
each_def([T|Ts],[proc(T/1,P)|Ps],Defs,Ss,M,D) :-
        def(T/1,P,D),
        body_defs(P,Ss,M,D),
        each_def(Ts,Ps,Defs,Ss,M,D).

body_defs([(_ :- B)|Cs],Ss,M,P) :-
        bodytypes(B,Ss,M,P),
        body_defs(Cs,Ss,M,P).
body_defs([],_,_,_).

bodytypes((B,Bs),Ss,M,P) :-
        !,
        functor(B,T,1),
        insertT(T,Ss,M),
        bodytypes(Bs,Ss,M,P).
bodytypes(true,_,_,_) :-
        !.
bodytypes(B,Ss,M,_) :-
        functor(B,T,1),
        insertT(T,Ss,M).


normalise_defs([],_,_,[]).
normalise_defs([proc(F,Cls)|Ps],T2,T,[proc(F,Cls1)|Ps1]) :-
        normalise(Cls,T2,T,Cls1),
        normalise_defs(Ps,T2,T,Ps1).


replaceproc(T1,[proc(T1/1,_)|Ps],Cls,[proc(T1/1,Cls)|Ps]) :-
        !.
replaceproc(T1,[P|Ps],Cls,[P|Ps1]) :-
        replaceproc(T1,Ps,Cls,Ps1).

normalise([], _,_,[]).
normalise([(H :- B)|Cls], T2,F,[(H :- B1)|T]) :-
        normalise_body(B,T2,F,B1),
        normalise(Cls,T2,F,T).

normalise_body(true,_,_,true) :-
        !.
normalise_body((B,Bs),T2,F,(B1,Bs1)) :-
        B =.. [T2,X],
        !,
        B1 =.. [F,X],
        normalise_body(Bs,T2,F,Bs1).
normalise_body((B,Bs),T2,F,(B,Bs1)) :-
        !,
        normalise_body(Bs,T2,F,Bs1).
normalise_body(B,T2,F,B1) :-
        B =.. [T2,X],
        !,
        B1 =.. [F,X].
normalise_body(B,_,_,B).

