/* regularise(T,P,D,N,N1):  D is set of RUL clauses in which defn of T 
   is not regular program.  P is RUL program. */

/* JPG: 16.1.92 */ 
/* JPG: 13.8.93 adapted from ub.pl */ 

%regularise(T,[proc(T/1,P1)|Subdefs],D,N,N2) :-
%	rul_defs(T/1,P,D),
%	merge_rul(P,P1,pairs([(T,T,T)|S],More),N,N1,D),
%	iterate_ub_merge(More,[(T,T,T)|S],Subdefs,N1,N2,D).

regularise(P,R,D,N,N1) :-
%	all_proc_names(D,S,M),
	iterate_ub_merge([(P,P,P)|_],[(P,P,P)|_],R,N,N1,D).

iterate_ub_merge([],_,[],N,N,_) :-
	!.
iterate_ub_merge(M,S,Defs,N,N2,D) :-
	each_ub_merge(M,Defs1, M1,S,N,N1,D),
	iterate_ub_merge(M1,S,Defs2,N1,N2,D),
	app(Defs1,Defs2,Defs).

each_ub_merge([],[],_,_,N,N,_) :-
	!.
each_ub_merge([(T,T,T)|Ps],[proc(T/1,Def)|Defs],More,S,N,N2,D) :-
	!,
	rul_defs(T/1,P1,D),
	merge_rul(P1,Def,pairs(S,More),N,N1,D),
	each_ub_merge(Ps,Defs,More,S,N1,N2,D).
each_ub_merge([(T1,T2,T3)|Ps],[proc(T3/1,Def)|Defs],More,S,N,N2,D) :-
	!,
	rul_defs(T1/1,P1,D),
	rul_defs(T2/1,P2,D),
	app(P1,P2,P3),
	renametypes(P3,T3,P4),
	merge_rul(P4,Def,pairs(S,More),N,N1,D),
	each_ub_merge(Ps,Defs,More,S,N1,N2,D).

merge_rul(P,Q,More,N,N1,D) :-
	remove((H :- B1),P,P1),
	removeclause(H,B1,P1,P2),	% two identical clauses 
	!,
	merge_rul([(H :- B1)|P2],Q,More,N,N1,D).
merge_rul(P,Q,More,N,N2,D) :-
	remove((H :- B1),P,P1),
	removeclause(H,B2,P1,P2),	% two clauses with same head 
	!,
	ub_body_merge(B1,B2,B,More,N,N1,D),
	merge_rul([(H :- B)|P2],Q,More,N1,N2,D).
merge_rul(P,P,_,N,N,_).

ub_body_merge(true,true,true,_,N,N,_) :-
	!.
ub_body_merge((B,Bs),(C,Cs),(D,Ds),More,N,N2,P) :-
	!,
	ub_atom_merge(B,C,D,More,N,N1,P),
	ub_body_merge(Bs,Cs,Ds,More,N1,N2,P).
ub_body_merge(B,C,D,More,N,N1,P) :-
	ub_atom_merge(B,C,D,More,N,N1,P).

ub_atom_merge(_,C,C,_,N,N,_) :-
	C =.. [any,_],
	!.
ub_atom_merge(B,_,B,_,N,N,_) :-
	B =.. [any,_],
	!.
ub_atom_merge(C,C,C,S,N,N1,_) :-
	!,
	insert_pair((C,C,C),S,N,N1).
ub_atom_merge(B,C,D,More,N,N1,_) :-
	B =.. [R,X],
	C =.. [S,X],
	get_name(R,S,T,More,N,N1),
	D =.. [T,X].

all_proc_names([proc(T/1,_)|Ps],S,M) :-
	memb1((T,T,T),S),
	memb1((T,T,T),M),
        all_proc_names(Ps,S,M).
all_proc_names([Qs|Ps],S,M) :-
        all_proc_names(Qs,S,M),
        all_proc_names(Ps,S,M).
all_proc_names([],_,_).

rul_defs(T/N,P,[proc(T/N,P1)|Ps]) :-
	!,
        rul_defs(T/N,P2,Ps),
	app(P1,P2,P).
rul_defs(T,P,[proc(_,_)|Ps]) :-
	!,
        rul_defs(T,P,Ps).
rul_defs(T,P,[Qs|Ps]) :-
        rul_defs(T,P1,Qs),
        rul_defs(T,P2,Ps),
	app(P1,P2,P).
rul_defs(_,[],[]).

renametypes([],_,[]) :-
	!.
renametypes([(H :- B)|Ps],T,[(S :- B)|Qs]) :-
	H =.. [_,X],
	S =.. [T,X],
	renametypes(Ps,T,Qs).

get_name(R,S,T,More,N,N1) :-
        R @=< S,
        !,
        insert_pair((R,S,T),More,N,N1).
get_name(R,S,T,More,N,N1) :-
        insert_pair((S,R,T),More,N,N1).


insert_pair(P,pairs(S,M),N,N1) :-
        insert_pair1(P,S,M,N,N1).

insert_pair1((A,A,A),S,M,N,N) :-
        var(S),
        !,
        S = [(A,A,A)|_],
        memb1((A,A,A),M).
insert_pair1((A,B,C),S,M,N,N1) :-
        var(S),
        !,
        newname(N,C),
        N1 is N+1,
        S = [(A,B,C)|_],
        memb1((A,B,C),M).
insert_pair1(P,[P|_],_,N,N) :-
        !.
insert_pair1(P,[_|S],M,N,N1) :-
        insert_pair1(P,S,M,N,N1).



