% Convert a program to a unary predicate program
% This is the first stage in producing a regular unary program

% JPG: 25.6.93


% unary(F,T,S):  F = source program file, T = list of program predicate types
%		 S = list of variable types

unary(F,T,S) :-
	see(F),
	read(C),
	unary_clauses(C,1,T,S,[]),
	seen,
	!.


unary_clauses(X,_,[],S,S) :- 
	end_of_file(X),
	!.
unary_clauses((A :- B),N,[(A :- Ts)|T],S,S2) :- 
	vars(A,Vs),
	canonical((A :- B)),
	var_types(Vs,B,N,N1,Ts),
	write_type_defs(Ts,B,S,S1),
	read(C1),
	unary_clauses(C1,N1,T,S1,S2).
unary_clauses(A,N,[(A :- Ts)|T],S,S2) :- 
	vars(A,Vs),
        var_types(Vs,true,N,N1,Ts),
        canonical((A :- Ts)),
        read(C1),
        unary_clauses(C1,N1,T,S,S2).


var_types([],_,N,N,true) :-
	!.
var_types([X],B,N,N1,S) :-
	make_body_atom(X,B,N,N1,S),
	!.
var_types([X|Xs],B,N,N2,(S,Ss)) :-
	make_body_atom(X,B,N,N1,S),
	var_types(Xs,B,N1,N2,Ss).
	
make_body_atom(X,B,N,N1,S) :-
	occurs_in_body(X,B),
	!,
	newtypename(N,P),
	S =.. [P,X],
	N1 is N+1.
make_body_atom(X,_,N,N,any(X)).

occurs_in_body(X,B) :-
	X == B,
	!.
occurs_in_body(X,B) :-
	nonvar(B),
	B =.. [_|Xs],
	occurs_in_body1(X,Xs).

occurs_in_body1(X,[Y|_]) :-
	occurs_in_body(X,Y),
	!.
occurs_in_body1(X,[_|Ys]) :-
	occurs_in_body1(X,Ys).

newtypename(K,P) :-
        name(K,Kn),
        name(s,T),
        app(T,Kn,Tn),
        name(P,Tn).


write_type_defs((any(_),Ts),B,S,S1) :-
	!,
	write_type_defs(Ts,B,S,S1).
write_type_defs((T,Ts),B,[(T :- B1)|S],S1) :-
	!,
	delete_non_occurs(T,B,B1),
	write_type_defs(Ts,B,S,S1).
write_type_defs(any(_),_,S,S)  :-
	!.
write_type_defs(T,B,[(T :- B1)|S],S) :-
	delete_non_occurs(T,B,B1).

delete_non_occurs(T,(B,Bs),(B,B1)) :-
	T =.. [_,X],
	occurs_in_body(X,B),
	!,	
	delete_non_occurs(T,Bs,B1).
delete_non_occurs(T,(_,Bs),B1) :-
	!,	
	delete_non_occurs(T,Bs,B1).
delete_non_occurs(T,B,B) :-
	T =.. [_,X],
	occurs_in_body(X,B),
	!.	
delete_non_occurs(_,_,true).


unary(F,P) :-
	unary(F,T,S),
	assert_type_clauses(T),
	bagof(Cl,[S]^unfold_clause(S,Cl),P),
	abolish(type_clause/2).

unfold_clause(S,(A :- Q)) :-
	memb2(C,S),
	melt(C,(A :- B)),
	unfold_each_lit(B,Q),
	canonical((A :- Q)).

unfold_each_lit((B,Bs),G) :-
	sp_builtin(B),
	!,
	unfold_each_lit(Bs,G).
unfold_each_lit((B,Bs),G) :-
	!,
	type_clause(B,K),
	unfold_each_lit(Bs,G1),
	joingoals(K,G1,G).
unfold_each_lit(B,true) :-
	sp_builtin(B),
	!.
unfold_each_lit(B,G) :-
	type_clause(B,G).

assert_type_clauses([]) :-
	!.
assert_type_clauses([Cl|Cls]) :-
	melt(Cl,(A :- B)),
	assert(type_clause(A,B)),
	assert_type_clauses(Cls).

write_cls(T) :-
	write('File name: '),
	read(F),
	tell(F),
	write_each_cl(T),
	told.


write_each_cl([]) :-
	!.
write_each_cl([C|Cs]) :-
	write(C),
	write('.'),
	nl,
	write_each_cl(Cs).
